/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *    Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-logger.c,v 1.1 1999/09/29 22:09:55 jordan Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff - undenyable logger stuff
 */

#include "common-stuff.h"

#ifdef USE_PTHREADS
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif
#endif

#ifdef HAVE_ASSERT_H
#include <assert.h>
#else
#define assert(x)
#endif

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#else
# include <socket.h> /* ????? */
#endif
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#include "hostnames.h"
#include "peks-logger.h"
#include "baseXX.h"
#include "crandom.h"
#include "cipher.h"
#include "rnd-pool.h"


/* ------------------------------------------------------------------------- *
 *                        private definitions                                *
 * ------------------------------------------------------------------------- */

#define   PAD '-' /* non space padding character, not in base64 charset */

typedef
struct _pub_flog {
  char  version  [1];	/* log record layout and contents idicator */
  char   cookie [10];	/* record identifier */
  char  prvhash [27];	/* allhash from the previos line */
  char  pubhash [27];	/* hash over public record fields */
  char  cfdhash [27];	/* hash over the confidential part */
} pub_flog ;


typedef
struct _conf_flog {
  char  logtime  [6];	/* time, when tis log was created */
  char   v6myip [11];	/* ip address of the logging server */
  char    myuid  [3];	/* user id of the logging server */
  char v6clntip [11];	/* client ip address */
  char clntuser  [9];	/* first part of client user name */
  char  v6trgip [11];	/* target ip of the system, the server addresses */
} conf_flog ;

typedef
struct _pub_log {
  pub_flog       fix ;
  const char *custom ;
} pub_log ;

typedef
struct _conf_log {
  conf_flog      fix ;
  const char *custom ;
} conf_log ;

typedef
struct _logset {
  pub_log         public ;
  conf_log  confidential ;
} logset ;

/* ------------------------------------------------------------------------- *
 *                        private variables                                  *
 * ------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------ *
 *                private functions: pthreads support                       *
 * ------------------------------------------------------------------------ */

#define SYSNAME "peks-logger"
#include "peks-sema.h"

/* ------------------------------------------------------------------------- *
 *                        private functions                                  *
 * ------------------------------------------------------------------------- */

#define buffer2ulong(b)    ( (((b) [0] << 24) & 0xff000000) |\
			     (((b) [1] << 16) & 0x00ff0000) |\
			     (((b) [2] <<  8) & 0x0000ff00) |\
		 	     (((b) [3]      ) & 0x000000ff) )

#define  ulong2buffer(b,u)  ( (b) [0] = ((u) >> 24) & 0xff ,\
                              (b) [1] = ((u) >> 16) & 0xff ,\
                              (b) [2] = ((u) >>  8) & 0xff ,\
                              (b) [3] = ((u)      ) & 0xff )

static void
store_string
 (char       *trg,
  unsigned   tlen,
  const char *str)
{
  int n ;
  
  if ((n = strlen (str)) > tlen) 
    n = tlen ;
  if (n > 0)
    memcpy (trg, str, n);
  if (n < tlen)
    memset (trg + n, PAD, tlen - n);
}


static void
store_ulong
 (char       *trg,
  unsigned   tlen,
  unsigned long u)
{
  char buf [10], *s ;

  ulong2buffer (buf, u);
  s = bin2base64 (buf, 4); /* ulong is 4, by definition */
  store_string (trg, tlen, s);
  XFREE (s);
}



static void
store_hash 
  (char           *trg,
   unsigned       tlen, 
   const char    *data,
   unsigned       size, 
   const char *xstring)
{
  static frame_desc *md ;
  const char *s ;
  int n;
  
  __enter_lock_semaphore ();

  if (md == 0) {
    md = create_frame (find_frame_class (PEKS_LOGGER_HASH_TYPE), 0);
    assert (md != 0);
  }
  
  XCRCFIRST (md, data, size);
  if (xstring != 0)
    XCRCNEXT (md, xstring, strlen (xstring));
  s = bin2base64 (XCRCRESULT0 (md), md->mdlen) ;

  __release_lock_semaphore ();

  store_string (trg, tlen, s);
  XFREE (s);
}



static int
finish_log_record
  (logset *l)
{
  char buf [sizeof (l->public.fix.cookie)] ;
  const char *s;
  struct in_addr in ;
  
  /* record the current version */
  l->public.fix.version [0] = PEKS_LOGGER_VERSION_CHAR ;

  /* store a base64 cookie */
  fast_random_bytes (buf, sizeof (buf)) ;
  s = bin2base64 (buf, sizeof (buf));
  store_string (l->public.fix.cookie, sizeof (l->public.fix.cookie), s);
  XFREE (s);

  /* store the current time */
  store_ulong (l->confidential.fix.logtime, 
	       sizeof (l->confidential.fix.logtime), time (0));
  
  /* store the current ip address */
  if ((s = get_host_ipaddress (0)) == 0 || inet_aton (s, &in) == 0)
    /* inet_aton () == 0 upon error, here */
    return -1;
  store_ulong (l->confidential.fix.v6myip, 
	       sizeof (l->confidential.fix.v6myip), in.s_addr);
  
  /* store the current user id */
  store_ulong (l->confidential.fix.myuid,
	       sizeof (l->confidential.fix.myuid), getuid());

  /* hash over the confidential part */
  store_hash (l->public.fix.cfdhash, sizeof (l->public.fix.cfdhash),
	      (char*)&l->confidential.fix, sizeof (l->confidential.fix),
	      l->confidential.custom);

  /* hash over the public part */
  store_hash (l->public.fix.pubhash, sizeof (l->public.fix.pubhash),
	      (char*)&l->public.fix, sizeof (l->public.fix),
	      l->public.custom);
  
  return 0 ;
}



static logset *
create_log_record
  (const char   *lasthash,
   const char   *clientip,
   const char *clientuser,
   const char      *trgip,
   const char *pub_custom,
   const char *cnf_custom)
{
  struct in_addr in ;
  logset *l = XMALLOC (sizeof (logset));
  
  /* store last hash */
  if (lasthash == 0) lasthash = "" ;
  store_string (l->public.fix.prvhash, 
		sizeof (l->public.fix.prvhash), lasthash);

  /* store client date */
  if (clientip == 0) in.s_addr = 0 ;
  else if (inet_aton (clientip, &in) == 0)
    /* inet_aton () == 0 upon error, here */
    return 0;
  store_ulong (l->confidential.fix.v6clntip, 
	       sizeof (l->confidential.fix.v6clntip), in.s_addr);
  if (clientuser == 0) clientuser = "" ;
  store_string (l->confidential.fix.clntuser, 
		sizeof (l->confidential.fix.clntuser), clientuser);

  /* store target address */
  if (trgip == 0) in.s_addr = 0 ;
  else if (inet_aton (trgip, &in) == 0)
    /* inet_aton () == 0 upon error, here */
    return 0;
  store_ulong (l->confidential.fix.v6trgip, 
	       sizeof (l->confidential.fix.v6trgip), in.s_addr);
  
  /* store custom data */
  l->public.custom       = pub_custom ;
  l->confidential.custom = cnf_custom ;

  /* add standard data */
  if (finish_log_record (l) == 0)
    return l;

  XFREE (l);
  return 0;
}

/* ------------------------------------------------------------------------- *
 *                         public functions                                  *
 * ------------------------------------------------------------------------- */

trans 
  ()
{
  logset *l = create_log_record (0,0,0,0,0,0);
}
